#! python3
import os
import sys
import numpy as np
import json
import pprint
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import argparse
import pandas as pd
import seaborn as sns
import plotting_config


class TimeseriesData:
    timestamps = []
    operations = []
    latencies = []
    scaleOutAverage = None



def plot_timeseries():

    parser = argparse.ArgumentParser(description='Plot timeseries and averages of different phases based on YCSB result files.')
    parser.add_argument('--runFolder', '-rf',type=str, nargs=1,required=True, help='path to the evalation_run_XXX folder')
    parser.add_argument('--scenario', '-s',type=str, nargs=1,required=True, choices=['elasticity', 'availability'], help='type of the evaluation scenario: elasticity/availability')
    parser.add_argument('--workloadPhase', '-wp',type=str, nargs=1,required=True, choices=['load', 'transaction'], help='type of the workload phase: load/transaction')
    parser.add_argument('--scaleDelay', '-sd',type=int, nargs='?',const=0, help='delay in seconds before scaling action was triggered')
    
    args = parser.parse_args()

    inputPath = args.runFolder[0]
    scenario = args.scenario[0]
    workloadPhase = args.workloadPhase[0]
    
    

    if scenario == 'elasticity':
        print("creating elasticity phase plots...")
        scaleDelay = args.scaleDelay

        deployVmDuration, scaleOutDuration = getPhaseDurations(inputPath)

        scaleOutStart = scaleDelay + deployVmDuration
        scaleOutEnd = scaleOutStart + scaleOutDuration

        plot_throughput_timeseries(inputPath,scaleOutStart,scaleOutEnd)

    elif scenario == 'availability':
        print("creating availability phase plots...")   
                
        df = getPhasesDataframe(inputPath)
        print(df)
        plot_availability_timeseries(inputPath,df,workloadPhase)


    print("finished evaluation metrics plotting!")

def plot_availability_timeseries(inputPath, df,workloadPhase):

    if workloadPhase == "load":
        dataPath = os.path.join(inputPath, plotting_config.DATA_FOLDER, plotting_config.DATA_LOAD_FILE)
        
    elif workloadPhase == "transaction":
        dataPath = os.path.join(inputPath, plotting_config.DATA_FOLDER, plotting_config.DATA_TRANSACTION_FILE)
    else:
        print("Exiting due to unsupported workloadPhase: " + workloadPhase)    
        sys.exit(1)

    extractedTimeseriesData = TimeseriesData()

    file = open(dataPath, 'r', encoding='utf8')
    try:
        for line in file:
            line = line.rstrip()
            if "[TIMESERIES]" in line:
                #extract current ops per second
                splittedLine = line.split(";")
                operationItem = splittedLine[2]
                operationItem = operationItem.strip()
                #print (operationItem)
                extractedTimeseriesData.operations.append(float(operationItem))
                #extract timestamp
                secondsItem = splittedLine[1]
                secondsItem = secondsItem.strip()
                extractedTimeseriesData.timestamps.append(float(secondsItem))

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")      


    initialAvg = calculate_phase_average(extractedTimeseriesData,0, df.loc[0,'InjectFailureEnd'])
    print("Initial (healthy) phase average is: " + str(initialAvg) + " ops/s")

    evaluationEnd = max(extractedTimeseriesData.timestamps)
   

    print("plotting throughput timeseries...")
    #Plotting 
    
    plt.figure()
    plt.xlabel("Runtime in sec")
    plt.ylabel("Operations / second")
    plt.title("Throughput")

    # Set axes range
    axes = plt.gca()    
    yMax = max(extractedTimeseriesData.operations) + 1000
    axes.set_ylim([0,yMax])
    
    plt.plot(extractedTimeseriesData.timestamps, extractedTimeseriesData.operations)  
    
    plt.hlines(y=initialAvg, xmin=0, xmax=(df.loc[0,'InjectFailureStart']), color='green', linestyles='dashed', zorder=3, label='healthy avg.')

    for index, row in df.iterrows():
        plt.axvline(df.loc[index,'InjectFailureStart'], color='red', linestyle=':', label="failure injection")
        #failure with no recovery following
        if  pd.isna(df.loc[index,'RecoveryVMStart']) and (index + 1) == len(df.index):
            unhealthyAvg = calculate_phase_average(extractedTimeseriesData,df.loc[index,'InjectFailureEnd'],evaluationEnd)

            plt.hlines(y=unhealthyAvg, xmin=df.loc[index,'InjectFailureEnd'], xmax=evaluationEnd, color='red', linestyles='dashed', zorder=3, label='unhealthy avg.')
        #failure with additional recovery 
        else:
            unhealthyAvg = calculate_phase_average(extractedTimeseriesData,df.loc[index,'InjectFailureEnd'],df.loc[index,'RecoveryDBMSStart'])

            plt.hlines(y=unhealthyAvg, xmin=df.loc[index,'InjectFailureEnd'], xmax=df.loc[index,'RecoveryDBMSStart'], color='red', linestyles='dashed', zorder=3, label='unhealthy avg.')

            #recovery avg
            recoveryAvg = calculate_phase_average(extractedTimeseriesData,df.loc[index,'RecoveryDBMSStart'],df.loc[index,'RecoveryDBMSEnd'])
            plt.hlines(y=recoveryAvg, xmin=df.loc[index,'RecoveryDBMSStart'], xmax=df.loc[index,'RecoveryDBMSEnd'], color='orange', linestyles='dashed', zorder=3, label='recovery avg.')
            #healthy avg
            if (index + 1) == len(df.index):
                healtyAvg = calculate_phase_average(extractedTimeseriesData,df.loc[index,'RecoveryDBMSEnd'],evaluationEnd)
                plt.hlines(y=healtyAvg, xmin=df.loc[index,'RecoveryDBMSEnd'], xmax=evaluationEnd, color='green', linestyles='dashed', zorder=3, label='healthy avg.')
            else:
                healtyAvg = calculate_phase_average(extractedTimeseriesData,df.loc[index,'RecoveryDBMSEnd'],df.loc[(index + 1),'InjectFailureEnd'])     
                plt.hlines(y=healtyAvg, xmin=df.loc[index,'RecoveryDBMSEnd'], xmax=df.loc[(index + 1),'InjectFailureEnd'], color='green', linestyles='dashed', zorder=3, label='healthy avg.')
            


    leg = plt.legend()

    outputFile = inputPath + plotting_config.TIMESERiES_FOLDER + "transaction-throughput-ts-phases.png"

    print("saving throughput plot to disk: " + outputFile)
    plt.savefig(outputFile)    

    outputDataframe = inputPath + plotting_config.TIMESERiES_FOLDER + "phases-dataframe"
    print("saving phases dataframe to disk: " + outputDataframe)
    df.to_pickle(outputDataframe)



def plot_throughput_timeseries(inputPath, scaleOutStart, scaleOutEnd):

    dataPath = inputPath + plotting_config.DATA_FOLDER + plotting_config.DATA_TRANSACTION_FILE

    extractedTimeseriesData = TimeseriesData()

    file = open(dataPath, 'r', encoding='utf8')
    try:
        for line in file:
            line = line.rstrip()
            if "[TIMESERIES]" in line:
                #extract current ops per second
                splittedLine = line.split(";")
                operationItem = splittedLine[2]
                operationItem = operationItem.strip()
                #print (operationItem)
                extractedTimeseriesData.operations.append(float(operationItem))
                #extract timestamp
                secondsItem = splittedLine[1]
                secondsItem = secondsItem.strip()
                extractedTimeseriesData.timestamps.append(float(secondsItem))

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")      


    initialAvg = calculate_phase_average(extractedTimeseriesData,0,scaleOutStart)
    print("Initial (overloaded) phase average is: " + str(initialAvg) + " ops/s")
    adaptationAvg = calculate_phase_average(extractedTimeseriesData,scaleOutStart,scaleOutEnd)
    print("Scale-out (overloaded) phase average is: " + str(adaptationAvg) + " ops/s")
    evaluationEnd = max(extractedTimeseriesData.timestamps)
    endAvg = calculate_phase_average(extractedTimeseriesData,scaleOutEnd,evaluationEnd)
    print("After scale-out (overloaded) phase average is: " + str(endAvg) + " ops/s")

    print("plotting throughput timeseries...")
    #Plotting 
    
    plt.figure()
    plt.xlabel("Runtime in sec")
    plt.ylabel("Operations / second")
    plt.title("Throughput")

    # Set axes range
    axes = plt.gca()    
    yMax = max(extractedTimeseriesData.operations) + 1000
    axes.set_ylim([0,yMax])
        
    plt.plot(extractedTimeseriesData.timestamps, extractedTimeseriesData.operations)   

    plt.hlines(y=initialAvg, xmin=0, xmax=scaleOutStart, color='red', linestyles='dashed', zorder=3, label='pre-scale-out avg.')

    plt.hlines(y=adaptationAvg, xmin=scaleOutStart, xmax=scaleOutEnd, color='orange', linestyles='dashed', zorder=3, label='scale-out avg.')

    plt.hlines(y=endAvg, xmin=scaleOutEnd, xmax=evaluationEnd, color='green', linestyles='dashed', zorder=3, label='post-scale-out avg.')

    plt.axvline(scaleOutStart, color='orange', linestyle=':', label="scale-out")

    leg = plt.legend()

    outputFile = inputPath + plotting_config.TIMESERiES_FOLDER + "transaction-throughput-ts-phases.png"

    print("saving throughput plot to disk: " + outputFile)
    plt.savefig(outputFile)   
   
 
def plot_latency_timeseries(inputPath, outputPath):

  
    extractedTimeseriesData = TimeseriesData()
    extractedTimeseriesData.timestamps = []


    file = open(inputPath, 'r', encoding='utf8')
    try:
        for line in file:
            line = line.rstrip()
            if "[TIMESERIES]" in line:
                #extract current ops per second
                splittedLine = line.split(";")
                latencyItem = splittedLine[3]
                latencyItem = latencyItem.strip() 
                #print (latencyItem)
                extractedTimeseriesData.latencies.append(float(latencyItem))
                #extract timestamp
                secondsItem = splittedLine[1]
                secondsItem = secondsItem.strip()
                extractedTimeseriesData.timestamps.append(float(secondsItem))

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")       

    print("plotting latency timeseries...")
    #Plotting 
    plt.figure()
    plt.xlabel("Runtime in sec")
    plt.ylabel("Avg. Latency in us")
    plt.title("Latency")
    plt.plot(extractedTimeseriesData.timestamps, extractedTimeseriesData.latencies)    

    outputFile = outputPath + "-latency-ts.png"
    print("saving latency plot to disk: " + outputFile)
    plt.savefig(outputFile)
    
    
def calculate_phase_average(extractedTimeseriesData, start,end):


    df = pd.DataFrame({'timestamps': extractedTimeseriesData.timestamps,'operations' : extractedTimeseriesData.operations})
    df = df[['timestamps','operations']]

    #print(df.to_string())

    #maximum = df['operations'].max()
    #print('Max value is: ' , maximum)
    #print("phase start: " + str(start) + " phase end: " + str(end))

    if(start > end):
        print("Start of current phase is after end of phase! Did the adapation finish after the transaction phase?")

    phaseDataFrame = df[df['timestamps'].between(start, end, inclusive=True)]
    #print(phaseDataFrame.to_string())

    average = phaseDataFrame['operations'].mean()
    #print('Phase average is: ' , average)

    return average


def getPhaseDurations(inputPath):

    #TODO: currently only supports one adaptation phase by using the second occurrence of the respective tasks!  

    inputPath = inputPath + plotting_config.TASK_FOLDER + plotting_config.TASK_LOG_FILE



    file = open(inputPath, 'r', encoding='utf8')

    deployVmStart = None
    deployVmFinish = None

    deployScaledDbStart = None
    deployScaledDbFinish = None

    try:
        for line in file:
            line = line.rstrip()
            if "SPAWN_VM" in line and "STARTED" in line:
                splittedLine = line.split(";")
                deployVmStart = splittedLine[1]
                deployVmStart = deployVmStart.replace('s','')
                deployVmStart = int(deployVmStart)
            if "SPAWN_VM" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                deployVmFinish = splittedLine[1]
                deployVmFinish = deployVmFinish.replace('s','')
                deployVmFinish = int(deployVmFinish) 
            if "DEPLOY_DB" in line and "STARTED" in line:
                splittedLine = line.split(";")
                deployScaledDbStart = splittedLine[1]
                deployScaledDbStart = deployScaledDbStart.replace('s','')
                deployScaledDbStart = int(deployScaledDbStart)
            if "DEPLOY_DB" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                deployScaledDbFinish = splittedLine[1]
                deployScaledDbFinish = deployScaledDbFinish.replace('s','')
                deployScaledDbFinish = int(deployScaledDbFinish)           

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")

      

    deployVmDuration = deployVmFinish - deployVmStart
    scaleOutDuration = deployScaledDbFinish - deployScaledDbStart 

    print("VM deploy duration is: " + str(deployVmDuration))
    print("scale-out  duration is: " + str(scaleOutDuration))  

    return deployVmDuration, scaleOutDuration


def getPhasesDataframe(inputPath):

    inputPath = inputPath + plotting_config.TASK_FOLDER + plotting_config.TASK_LOG_FILE

    file = open(inputPath, 'r', encoding='utf8')

    colNames =['YCSBLoadStart','YCSBTransactionStart','GibbonStart','GibbonEnd', 'InjectFailureStart', 'InjectFailureEnd','RecoveryVMStart','RecoveryVMEnd','RecoveryDBMSStart','RecoveryDBMSEnd']
    df = pd.DataFrame(columns = colNames)
    
    indexCounter = 0
    try:
        for line in file:
            line = line.rstrip()
            if "YCSB_LOAD" in line and "STARTED" in line:
                splittedLine = line.split(";")
                ycsbLoadStart = splittedLine[1]
                ycsbLoadStart = ycsbLoadStart.replace('s','')
                ycsbLoadStart = int(ycsbLoadStart)
                df.loc[indexCounter,'YCSBLoadStart'] = ycsbLoadStart

            if "YCSB_TRANSACTION" in line and "STARTED" in line:
                splittedLine = line.split(";")
                ycsbTransactionStart = splittedLine[1]
                ycsbTransactionStart = ycsbTransactionStart.replace('s','')
                ycsbTransactionStart = int(ycsbTransactionStart)
                df.loc[indexCounter,'YCSBTransactionStart'] = ycsbTransactionStart

            if "GIBBON" in line and "STARTED" in line:
                splittedLine = line.split(";")
                GibbonStart = splittedLine[1]
                GibbonStart = GibbonStart.replace('s','')
                GibbonStart = int(GibbonStart)
                df.loc[indexCounter,'GibbonStart'] = GibbonStart    

            if "INJECT_FAILURE" in line and "STARTED" in line:
                splittedLine = line.split(";")
                InjectStart = splittedLine[1]
                InjectStart = InjectStart.replace('s','')
                InjectStart = int(InjectStart)
                df.loc[indexCounter,'InjectFailureStart'] = InjectStart 
             
            if "INJECT_FAILURE" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                InjectEnd = splittedLine[1]
                InjectEnd = InjectEnd.replace('s','')
                InjectEnd = int(InjectEnd)
                df.loc[indexCounter,'InjectFailureEnd'] = InjectEnd   
            
            if "RECOVERY_SPAWN_VM" in line and "STARTED" in line:
                splittedLine = line.split(";")
                revoveryVMStart = splittedLine[1]
                revoveryVMStart = revoveryVMStart.replace('s','')
                revoveryVMStart = int(revoveryVMStart)
                df.loc[indexCounter,'RecoveryVMStart'] = revoveryVMStart    


            if "RECOVERY_SPAWN_VM" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                revoveryVMEnd = splittedLine[1]
                revoveryVMEnd = revoveryVMEnd.replace('s','')
                revoveryVMEnd = int(revoveryVMEnd)
                df.loc[indexCounter,'RecoveryVMEnd'] = revoveryVMEnd  
            
            if "RECOVERY_DEPLOY_DB" in line and "STARTED" in line:
                splittedLine = line.split(";")
                revoveryDBMSStart = splittedLine[1]
                revoveryDBMSStart = revoveryDBMSStart.replace('s','')
                revoveryDBMSStart = int(revoveryDBMSStart)
                df.loc[indexCounter,'RecoveryDBMSStart'] = revoveryDBMSStart    


            if "RECOVERY_DEPLOY_DB" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                revoveryDBMSEnd = splittedLine[1]
                revoveryDBMSEnd = revoveryDBMSEnd.replace('s','')
                revoveryDBMSEnd = int(revoveryDBMSEnd)
                df.loc[indexCounter,'RecoveryDBMSEnd'] = revoveryDBMSEnd      
             
            if "GIBBON" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                GibbonEnd = splittedLine[1]
                GibbonEnd = GibbonEnd.replace('s','')
                GibbonEnd = int(GibbonEnd)
                df.loc[indexCounter,'GibbonEnd'] = GibbonEnd    
                indexCounter = indexCounter + 1    
                
                
            
    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")

    print("non-normalized data frame:")
    print(df)

    # normalize dataframe for load phase
    if pd.isna(df.loc[0,'YCSBTransactionStart']):
        print("normalizing for load phase")
        df.GibbonStart = df.GibbonStart - df.loc[0,'YCSBLoadStart']
        df.GibbonEnd = df.GibbonEnd - df.loc[0,'YCSBLoadStart']
        df.InjectFailureStart = df.InjectFailureStart - df.loc[0,'YCSBLoadStart']
        df.InjectFailureEnd = df.InjectFailureEnd - df.loc[0,'YCSBLoadStart']
        df.RecoveryVMStart = df.RecoveryVMStart - df.loc[0,'YCSBLoadStart']
        df.RecoveryVMEnd = df.RecoveryVMEnd - df.loc[0,'YCSBLoadStart']
        df.RecoveryDBMSStart = df.RecoveryDBMSStart - df.loc[0,'YCSBLoadStart']
        df.RecoveryDBMSEnd = df.RecoveryDBMSEnd - df.loc[0,'YCSBLoadStart']

    # normalize dataframe for transaction phase    
    else:     
        print("normalizing for transaction phase")
        df.GibbonStart = df.GibbonStart - df.loc[0,'YCSBTransactionStart']
        df.GibbonEnd = df.GibbonEnd - df.loc[0,'YCSBTransactionStart']
        df.InjectFailureStart = df.InjectFailureStart - df.loc[0,'YCSBTransactionStart']
        df.InjectFailureEnd = df.InjectFailureEnd - df.loc[0,'YCSBTransactionStart']
        df.RecoveryVMStart = df.RecoveryVMStart - df.loc[0,'YCSBTransactionStart']
        df.RecoveryVMEnd = df.RecoveryVMEnd - df.loc[0,'YCSBTransactionStart']
        df.RecoveryDBMSStart = df.RecoveryDBMSStart - df.loc[0,'YCSBTransactionStart']
        df.RecoveryDBMSEnd = df.RecoveryDBMSEnd - df.loc[0,'YCSBTransactionStart']
        
   

    return df


#init function call
plot_timeseries()