import argparse
import glob
import os
import fnmatch

import matplotlib
import pandas as pd

matplotlib.use('Agg')
from matplotlib import pyplot as plt

import plotting_config


class TimeSeriesAggregator:
    """
    Main object of the class. Stores many relevant values and data frames used for plotting
    """
    def __init__(self):
        self.parser = None
        self.input_path = None
        self.output_path = None
        self.agg_latency = None
        self.agg_throughput = None
        self.cols_latency = None
        self.cols_throughput = None
        self.run = -1
        self.threads = 1
        self.files = 0
        self.min_t = 0
        self.max_t = 0
        self.min_l = 0
        self.max_l = 0

    def parser_setup(self):
        """
        sets up a parser that allows to read parameters when executing
        """

        #TODO: add output format as parameter: pdf/png and align plotting call

        # define input directory with --input/-i and the output path with --output/-o on which
        # the resulting plots shall be stored
        parser = argparse.ArgumentParser(description='Plot timeseries based on YCSB result files.')
        parser.add_argument('--input', '-i', type=str, nargs=1, required=True,
                            help='path of the folder with the load.txt or transaction.txt')
        #parser.add_argument('--output', '-o', type=str, nargs=1, required=True,
        #                    help='path to the store timeseries plot')
        
       
        # optional argument: run number (the specified run will be shown in the whole context)
        parser.add_argument('--threads', '-t', type=int, choices=range(1, 129), required=True)
        # fix the parser as main parser object of the TimeSeriesAggregator object
        self.parser = parser

    def parse_input(self):
        # process input parameters
        args = self.parser.parse_args()
        # define inputPath as the directory with load.txt files
        self.input_path = args.input[0]          
        self.output_path = self.input_path
       

        # derive the number of runs based on the number of *.txt files in the folder
        self.run = len(fnmatch.filter(os.listdir(self.input_path), '*.txt'))
       
        self.threads = args.threads



    def extract_input(self):
        """
        data is prepared and processed for later plotting
        """
        # get all files from the given directory
        all_files = glob.glob(os.path.join(self.input_path, "*.txt"))
        file_list = []
        # attribute that determines how many timestamp-rows will be created
        current_max = 0
        # create one data frame for each read file
        for filename in all_files:
            df = pd.read_csv(filename, error_bad_lines=False, sep=";",
                             names=['timeseries', 'timestamp', 'throughput', 'latency'],
                             index_col=False)
            file_list.append(df)
            self.files = len(file_list)

            # determine the current maximum of timestamps (fixes the length of the data frame later)
            maximum_timestamp_of_file = max(df['timestamp'])
            if maximum_timestamp_of_file > current_max:
                current_max = maximum_timestamp_of_file

        # create an empty data frame with only timestamps for latency and throughput
        self.cols_latency = ['timestamp'] + ['l{}'.format(i) for i in range(len(all_files))]
        self.cols_throughput = ['timestamp'] + ['t{}'.format(i) for i in range(len(all_files))]
        agg_frame_latency = pd.DataFrame(columns=self.cols_latency)
        agg_frame_latency['timestamp'] = range(10, int(current_max) + 10, 10)
        agg_frame_throughput = pd.DataFrame(columns=self.cols_throughput)
        agg_frame_throughput['timestamp'] = range(10, int(current_max) + 10, 10)

        # fill the new data frames, enumerate columns with file numbers
        for index, file in enumerate(file_list):
            agg_frame_latency['l{}'.format(index)] = file['latency']
            agg_frame_throughput['t{}'.format(index)] = file['throughput']

        # change the index so that timestamps are the index
        agg_frame_latency.set_index('timestamp', inplace=True)
        agg_frame_throughput.set_index('timestamp', inplace=True)
        # calculate mean for each row
        agg_frame_latency['Mean'] = agg_frame_latency.mean(axis=1, skipna=True)
        agg_frame_throughput['Mean'] = agg_frame_throughput.mean(axis=1, skipna=True)

        indices_latency = [x for x in agg_frame_latency.columns if x != "Mean"]
        indices_throughput = [x for x in agg_frame_throughput.columns if x != "Mean"]

        # the factor of the standard deviation is 1/n-1, thus it is clearly specified as ddof=1
        # calculate standard deviation per row
        agg_frame_latency['Standard Deviation'] = agg_frame_latency[indices_latency].std(axis=1, skipna=True, ddof=1)
        agg_frame_throughput['Standard Deviation'] = agg_frame_throughput[indices_throughput].std(axis=1, skipna=True, ddof=1)
        # save results in self-object
        self.agg_latency = agg_frame_latency
        self.agg_throughput = agg_frame_throughput

    def extract_dataframe(self):
        """
        processes the collected data and generates plots out of the given data
        """

        # THROUGHPUT
        # indicate the start of the plotting process
        print("Extracting timeseries data...")

        self.min_t = self.agg_throughput.iloc[:, :-2].min().min()
        self.max_t = self.agg_throughput.iloc[:, :-2].max().max()

        timestamps_t = self.agg_throughput.index
        filename = str(self.threads) + "-threads"
        output = os.path.join(self.output_path,filename)

        print(self.agg_throughput)
        self.agg_throughput.to_pickle(output)

    def process(self):
        """
        main method. Executes the methods in correct order and terminates after running
        """
        self.parser_setup()
        self.parse_input()
        self.extract_input()
        self.extract_dataframe()


if __name__ == "__main__":
    TimeSeriesAggregator().process()
