package de.uulm.omi.evaluation.task;

import static com.google.common.base.Preconditions.checkNotNull;

import com.github.rholder.retry.RetryException;
import com.github.rholder.retry.Retryer;
import com.github.rholder.retry.RetryerBuilder;
import com.github.rholder.retry.StopStrategies;
import com.github.rholder.retry.WaitStrategies;
import com.google.common.base.Predicates;
import de.uniulm.omi.cloudiator.colosseum.client.Client;
import de.uniulm.omi.cloudiator.colosseum.client.entities.ApplicationComponent;
import de.uniulm.omi.cloudiator.colosseum.client.entities.Instance;
import de.uniulm.omi.cloudiator.colosseum.client.entities.LifecycleComponent;
import de.uniulm.omi.cloudiator.colosseum.client.entities.VirtualMachine;
import de.uniulm.omi.cloudiator.colosseum.client.entities.enums.RemoteState;
import de.uulm.omi.cloudiator.dbms.evaluator.model.ClusterContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.ElasticityStep.ScalingTypeEnum;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.GibbonSpec;
import de.uulm.omi.evaluation.jungle.GibbonAlgorithm;
import de.uulm.omi.utils.ColosseumClientHelper;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import org.jeasy.flows.work.DefaultWorkReport;
import org.jeasy.flows.work.WorkReport;
import org.jeasy.flows.work.WorkStatus;

/**
 * Created by Daniel Seybold on 12.08.2019.
 */
public class GibbonV2 extends EvaluationTask {


  private final Client colosseumClient;

  private final List<GibbonSpec> gibbonSpecs;
  private final ClusterContext clusterContext;
  private String loggerPrefix;

  private final long VM_CREATION_WAITING_PERIOD = 10000;
  private final long INSTANCE_CREATION_WAITING_PERIOD = 20000;
  private final ScalingTypeEnum scalingType =  ScalingTypeEnum.OUT;

  public GibbonV2(EvaluationContext evaluationContext, ClusterContext dbmsClusterContext, List<GibbonSpec> gibbonSpecs) {
    super(TaskType.GIBBON, evaluationContext);

    this.gibbonSpecs = gibbonSpecs;
    this.clusterContext = dbmsClusterContext;
    this.colosseumClient = ColosseumClientHelper.getColosseumClient();
  }


  @Override
  public WorkReport call() {

    this.loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";


    Long appInstanceId = Long.valueOf(this.clusterContext.getIdApplicationInstance());

    checkNotNull(appInstanceId);
    LOGGER.debug(loggerPrefix + "Executing GibbonAlgorithm for applicationInstanceId: " + appInstanceId);

    /*
    Execute all Gibbon Specs sequentially
     */
    int counter = 0;
    boolean check = false;
    for (GibbonSpec gibbonSpec: this.gibbonSpecs ) {
      new LogTask(evaluationContext, TaskType.GIBBON, TaskState.STARTED).call();

      //1. log GibbonSpec to disk
      writeGibbonSpec(counter,gibbonSpec);

      //2. wait specified delay before injecting failure
      check = this.gibbonDelayWait(gibbonSpec.getGibbonDelay());
      if(!check){
        return new DefaultWorkReport(WorkStatus.FAILED);
      }

      //3. inject failure
      new LogTask(evaluationContext, TaskType.INJECT_FAILURE, TaskState.STARTED).call();
      try {
        GibbonAlgorithm.injectFailure(appInstanceId, gibbonSpec);
      }catch (Exception e){
        LOGGER.error(loggerPrefix + "error while executing the GibbonAlgorithm algorithm!" ,e);
        return new DefaultWorkReport(WorkStatus.FAILED);
      }
      new LogTask(evaluationContext, TaskType.INJECT_FAILURE, TaskState.FINISHED).call();

      //4. wait for recoveryDelay
      check = this.gibbonDelayWait(gibbonSpec.getRecoveryDelay());
      if(!check){
        LOGGER.debug(loggerPrefix + " wait before recovery failed for applicationInstanceId: " + appInstanceId);
        return new DefaultWorkReport(WorkStatus.FAILED);
      }

      //5. execute revoery if specified
      if(gibbonSpec.isFailureRecovery()){

        check = this.recovery(appInstanceId);
        if(!check){
          LOGGER.debug(loggerPrefix + " Executing recovery failed for applicationInstanceId: " + appInstanceId);
          return new DefaultWorkReport(WorkStatus.FAILED);
        }

      }

      counter++;
      new LogTask(evaluationContext, TaskType.GIBBON, TaskState.FINISHED).call();
    }



    return new DefaultWorkReport(WorkStatus.COMPLETED);
  }


  private boolean recovery(Long appInstanceId){

    new LogTask(evaluationContext, TaskType.RECOVERY, TaskState.STARTED).call();

    List<Instance> instances = colosseumClient.controller(Instance.class).getList();
    Instance scaleOutTemplateInstance = null;

    //Search for matching DATA component VM
    for (Instance instance : instances){
      if(instance.getApplicationInstance().equals(appInstanceId)){

        //check for DATA component
        ApplicationComponent applicationComponent = colosseumClient.controller(ApplicationComponent.class).get(instance.getApplicationComponent());

        LifecycleComponent component = colosseumClient.controller(LifecycleComponent.class)
            .get(applicationComponent.getComponent());

        if(component.getName().equals("DATA")){
          scaleOutTemplateInstance = instance;
          break;
        }

      }

    }

    if(scaleOutTemplateInstance == null){
      LOGGER.warn(loggerPrefix + "No matching DATA component found applicationInstanceId: " + appInstanceId + "! Cancelling scale-out!");
      return false;

    }else {
      LOGGER.debug(loggerPrefix + "Found  matching DATA component with applicationInstanceId: " + appInstanceId );
    }

    //scale-out VM
    VirtualMachine scaleOutVm = this.scaleOutVM(scaleOutTemplateInstance);
    LOGGER.debug(loggerPrefix + "Successfully sent create VM request!");



    //scale-out Instance
    Instance scaleOutInstance = this.scaleOutInstance(scaleOutTemplateInstance, scaleOutVm);
    LOGGER.debug(loggerPrefix + "Successfully sent create Instance request!");


    //wait until VM is created
    LOGGER.debug(loggerPrefix + "Waiting for VM creation");
    new LogTask(evaluationContext, TaskType.RECOVERY_SPAWN_VM, TaskState.STARTED).call();
    if(!checkIfVmIsReady(scaleOutVm)){
      LOGGER.debug(loggerPrefix + "Create scale-out VM failed!");
      return false;
    }
    new LogTask(evaluationContext, TaskType.RECOVERY_SPAWN_VM, TaskState.FINISHED).call();


    //wait until DBMS is deployed on VM
    LOGGER.debug(loggerPrefix + "Waiting for Instance creation!");
    new LogTask(evaluationContext, TaskType.RECOVERY_DEPLOY_DB, TaskState.STARTED).call();
    waitUntilInstanceIsReady(scaleOutInstance);
    new LogTask(evaluationContext, TaskType.RECOVERY_DEPLOY_DB, TaskState.FINISHED).call();


    LOGGER.info(loggerPrefix + "Scale-out successful, created new component instance with ID: " + scaleOutInstance.getId());
    new LogTask(evaluationContext, TaskType.RECOVERY, TaskState.FINISHED).call();

    return true;
  }



  private VirtualMachine scaleOutVM(Instance scaleOutTemplateInstance){

    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";


    Long vmId = scaleOutTemplateInstance.getVirtualMachine();

    VirtualMachine virtualMachine = colosseumClient.controller(VirtualMachine.class).get(vmId);

    VirtualMachine scaleOutVm = new VirtualMachine(
        null,
        null,
        null,
        null,
        virtualMachine.getCloud(),
        virtualMachine.getCloudCredentials(),
        virtualMachine.getOwner(),
        virtualMachine.getLocation(),
        virtualMachine.getName() + "ScaleOut_" + System.currentTimeMillis(),
        virtualMachine.getImage(),
        virtualMachine.getHardware(),
        virtualMachine.getTemplateOptions()
    );

    LOGGER.debug(loggerPrefix + "Creating new VM for scale-out...");
    scaleOutVm = colosseumClient.controller(VirtualMachine.class).create(scaleOutVm);

    return scaleOutVm;
  }

  private Instance scaleOutInstance(Instance scaleOutTemplateInstance, VirtualMachine scaleOutVm){

    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";

    Instance scaleOutInstance = new Instance(
        null,
        null,
        null,
        null,
        scaleOutTemplateInstance.getApplicationComponent(),
        scaleOutTemplateInstance.getApplicationInstance(),
        scaleOutVm.getId()
    );

    LOGGER.debug(loggerPrefix + "Creating new DATA instance for scale-out...");
    scaleOutInstance = colosseumClient.controller(Instance.class).create(scaleOutInstance);

    return scaleOutInstance;
  }

  private boolean checkIfVmIsReady(VirtualMachine virtualMachine){
    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";
    while (true){
      VirtualMachine createdVM = colosseumClient.controller(VirtualMachine.class).get(virtualMachine.getId());
      if(createdVM.getRemoteState() != null && createdVM.getRemoteState().equals(RemoteState.ERROR)){
        return false;
      }else if(createdVM.getRemoteState() != null && createdVM.getRemoteState().equals(RemoteState.OK)){
        return true;
      }

      LOGGER.debug(loggerPrefix + "VM creating still in progress, waiting for " + VM_CREATION_WAITING_PERIOD/1000 + " seconds!");
      try {

        Thread.sleep(VM_CREATION_WAITING_PERIOD);

      } catch (InterruptedException e) {
        LOGGER.error(loggerPrefix + "Thread got interrupted while waiting for creating VM...", e);
      }
    }

  }

  private void waitUntilInstanceIsReady(Instance instance){

    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";

    Callable<Boolean> checkInstanceTask = new Callable<Boolean>() {
      public Boolean call() throws Exception {
        Instance createdInstance = colosseumClient.controller(Instance.class).get(instance.getId());

        if(createdInstance != null && createdInstance.getRemoteState() != null &&  createdInstance.getRemoteState().equals(RemoteState.ERROR)){
          LOGGER.error(loggerPrefix + "Instance resulted in error state, aborting!");
          return false;

        }else if(createdInstance != null && createdInstance.getRemoteState() != null && createdInstance.getRemoteState().equals(RemoteState.OK)){
          LOGGER.info(loggerPrefix + "Instance successfully created and in OK state!");
          return true;
        }

        LOGGER.debug(loggerPrefix + "Instance creating still in progress, waiting for " + INSTANCE_CREATION_WAITING_PERIOD/1000 + " seconds!");
        return false;
      }
    };

    Retryer<Boolean> retryer = RetryerBuilder.<Boolean>newBuilder()
        .retryIfResult(Predicates.<Boolean>equalTo(false))
        .retryIfExceptionOfType(IOException.class)
        .withWaitStrategy(WaitStrategies.fixedWait(10, TimeUnit.SECONDS))
        .withStopStrategy(StopStrategies.neverStop())
        .build();

    try {
      retryer.call(checkInstanceTask);
    } catch (RetryException e) {
      LOGGER.error(loggerPrefix + e.getMessage() );
      e.printStackTrace();
    } catch (ExecutionException e) {
      LOGGER.error(loggerPrefix + e.getMessage() );
      e.printStackTrace();
    }
  }

  private void writeGibbonSpec(int counter, GibbonSpec gibbonSpec){
    String specUri =  this.evaluationContext.getEvluationRun()
        + File.separator + EvaluationTask.SPECS_FOLDER
        + File.separator + counter + "_" + EvaluationTask.GIBBON_SPEC;
    this.writeReport(specUri, GSON.toJson(gibbonSpec));
  }

  private boolean gibbonDelayWait(long delay){

    try {
      long delayInMilliseconds = delay * 1000;
      LOGGER.debug(loggerPrefix + "Going to sleep for: " + delay + " seconds");
      Thread.sleep(delayInMilliseconds);
      return true;
    } catch (InterruptedException e) {
      LOGGER.error(loggerPrefix + "Delay Thread got interrupted!",e);
      e.printStackTrace();
      return false;
    }

  }
}