package de.uulm.omi.evaluation.task;

import com.google.common.io.CharStreams;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import de.uniulm.omi.cloudiator.colosseum.client.Client;
import de.uniulm.omi.cloudiator.colosseum.client.entities.KeyPair;
import de.uulm.omi.cloudiator.dbms.evaluator.model.ClusterContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationContext;
import de.uulm.omi.utils.ColosseumClientHelper;
import de.uulm.omi.utils.ColosseumUtils;
import de.uulm.omi.utils.DBMSNodeType;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import org.jeasy.flows.work.DefaultWorkReport;
import org.jeasy.flows.work.WorkReport;
import org.jeasy.flows.work.WorkStatus;

/**
 * Created by Daniel Seybold on 19.08.2019.
 */
public class RemoteCommand extends EvaluationTask {

  private final Client colosseumClient;
  private String loggerPrefix;
  private final String DBMSType;

  private final ClusterContext clusterContext;
  public RemoteCommand(EvaluationContext evaluationContext, ClusterContext dbmsClusterContext, String DBMSType) {
    super(TaskType.REMOTE_COMMAND, evaluationContext);

    this.colosseumClient = ColosseumClientHelper.getColosseumClient();

    this.clusterContext = dbmsClusterContext;
    this.DBMSType = DBMSType;
  }

  @Override
  public WorkReport call() {

    this.loggerPrefix = "RUN " + this.evaluationContext.getRunCounter().intValue() +": ";

    Long appInstanceId = Long.valueOf(this.clusterContext.getIdApplicationInstance());

    List<String> endpoints = ColosseumUtils
        .getEndpoints(appInstanceId, "PUBLIC", DBMSNodeType.SEED);

    if (endpoints.isEmpty() || endpoints.size()>1){
      LOGGER.error( this.loggerPrefix + "Invalid number of Seed endpoints: " + endpoints.size());
      return new DefaultWorkReport(WorkStatus.FAILED);
    }

    List<String> privateEndpoints = ColosseumUtils
        .getEndpoints(appInstanceId, "PRIVATE", DBMSNodeType.SEED);

    if (privateEndpoints.isEmpty() || privateEndpoints.size()>1){
      LOGGER.error( this.loggerPrefix + "Invalid number of Seed private endpoints: " + privateEndpoints.size());
      return new DefaultWorkReport(WorkStatus.FAILED);
    }

    String seedEnpoint = endpoints.get(0);
    String seedPrivateEndpoint = privateEndpoints.get(0);
    String recordCountCommand = null;
    String statusCommand = null;
    if(this.DBMSType.toLowerCase().equals("cassandra")){
      recordCountCommand = getCassandraRecordCountCommand(seedPrivateEndpoint);
      statusCommand = getCassandraClusterStateCommand();

    }else if(this.DBMSType.toLowerCase().equals("couchbase")){
      recordCountCommand = getCouchbaseRecordCountCommand();
      statusCommand = getCouchbaseClusterStateCommand();
    }else {
      LOGGER.error(this.loggerPrefix + "Unsupported command target: " + this.DBMSType + "! Skipping RemoteCommand Task!");
      return new DefaultWorkReport(WorkStatus.COMPLETED);
    }

    /**
     * get correct key pair
     */
    //TODO: derive idCloud dynamically depending on application
    long idCloud = 1;
    String privateKeyPlain = null;
    String publicKeyPlain = null;


    List<KeyPair> keyPairs = colosseumClient.controller(KeyPair.class).getList();
    for(KeyPair keyPair : keyPairs){
      if(keyPair.getCloud().equals(idCloud)){
        LOGGER.debug(this.loggerPrefix + "Found matching keypair for cloud with id: " + keyPair.getId());
        privateKeyPlain = keyPair.getPrivateKey();
        publicKeyPlain = keyPair.getPublicKey();

      }
    }

    /**
     * SSH connection part
     */

    //TODO: refactor into a generic ssh command helper tool and make this task the getNumberOfRecords task



    final String user = "ubuntu";
    final int port = 22;


    try {

      /**
       * get cluster state session
       */
      JSch jsch = new JSch();
      jsch.addIdentity("test_rsa", privateKeyPlain.getBytes(), publicKeyPlain.getBytes(), "secret".getBytes());

      Session session = jsch.getSession(user, seedEnpoint, port);
      session.setConfig("StrictHostKeyChecking", "no");
      LOGGER.debug("session created.");
      session.connect();

      ChannelExec channel = (ChannelExec) session.openChannel("exec");
      channel.setInputStream(null);
      InputStream output = channel.getInputStream();
      channel.setErrStream(System.err);
      channel.setCommand(statusCommand);

      channel.connect();
      String clusterStateResult = CharStreams.toString(new InputStreamReader(output));
      channel.disconnect();

      LOGGER.debug(this.loggerPrefix +  "The remote clusterState command is: " + clusterStateResult);

      session.disconnect();


      /**
       * Log the cluster state
       */
      logCustomEntry("CLUSTER_STATE",clusterStateResult);

      JSch jsch2 = new JSch();
      jsch2.addIdentity("test_rsa", privateKeyPlain.getBytes(), publicKeyPlain.getBytes(), "secret".getBytes());

      /**
       * get record count session
       */
      Session session2 = jsch2.getSession(user, seedEnpoint, port);
      session2.setConfig("StrictHostKeyChecking", "no");
      session2.connect();
      LOGGER.debug("session2 connected!");


      ChannelExec channel2 = (ChannelExec) session2.openChannel("exec");
      channel2.setCommand(recordCountCommand);


      /*
      channel2.setInputStream(null);
      InputStream output2 = channel2.getInputStream();
      channel2.setErrStream(System.err);
      LOGGER.debug(recordCountCommand);
      channel2.setCommand(recordCountCommand);

      channel2.connect(180000);
      String recordCountResult = CharStreams.toString(new InputStreamReader(output2));
      channel2.disconnect();


      LOGGER.debug(this.loggerPrefix +  "The remote recordCount command result is: " + recordCountResult);

      session2.disconnect();


       */


      InputStream commandOutput = channel2.getExtInputStream();

      StringBuilder outputBuffer = new StringBuilder();
      StringBuilder errorBuffer = new StringBuilder();

      InputStream in = channel2.getInputStream();
      InputStream err = channel2.getExtInputStream();

      channel2.connect();

      byte[] tmp = new byte[1024];
      while (true) {
        while (in.available() > 0) {
          int i = in.read(tmp, 0, 1024);
          if (i < 0) break;
          outputBuffer.append(new String(tmp, 0, i));
        }
        while (err.available() > 0) {
          int i = err.read(tmp, 0, 1024);
          if (i < 0) break;
          errorBuffer.append(new String(tmp, 0, i));
        }
        if (channel2.isClosed()) {
          if ((in.available() > 0) || (err.available() > 0)) continue;
          LOGGER.debug(this.loggerPrefix +  "The remote recordCount command exit-status: " + channel2.getExitStatus());
          break;
        }
        try {
          Thread.sleep(1000);
        } catch (Exception ee) {
          LOGGER.debug("Exception while executing record count command!",ee);
        }
      }

      String recordCountResult = outputBuffer.toString();
      String recordCountError = errorBuffer.toString();

      LOGGER.debug(this.loggerPrefix +  "The remote recordCount command result is: " + recordCountResult);
      LOGGER.debug(this.loggerPrefix +  "The remote recordCount error  result is: " + recordCountError);


      channel2.disconnect();

      /**
       * Log the total number of records
       */
     logCustomEntry("RECORD_COUNT",recordCountResult);



    } catch (JSchException e) {
      LOGGER.error("Error while executing SSH command!",e);
      return new DefaultWorkReport(WorkStatus.FAILED);

    } catch (IOException e) {
      LOGGER.error("Error while reading command result!",e);
      return new DefaultWorkReport(WorkStatus.FAILED);
    }


    return new DefaultWorkReport(WorkStatus.COMPLETED);
  }

  private void logCustomEntry(String id, String value){

    long currentTimestamp = System.currentTimeMillis()/1000;

    long eventTimestamp = currentTimestamp - this.evaluationContext.getStartTime();

    String taskLogMessage = new SimpleDateFormat("dd.MM.yyyy-HH:mm:ss").format(new Date()) +"; " +eventTimestamp + "s; " + id + "; " + value + "\n";


    String taskLogUri =  this.evaluationContext.getEvluationRun()
        + File.separator + EvaluationTask.TASK_LOG_FOLDER
        + File.separator + EvaluationTask.TASK_LOG;



    try {
      Files.write( Paths.get(taskLogUri), taskLogMessage.getBytes(), StandardOpenOption.APPEND);
    } catch (IOException e) {
      LOGGER.error("Error while writing TaskLog!" , e);
    }
  }


  private static String getCassandraRecordCountCommand(String host){

    //TODO: check if these values need to be derived dynamically based on total number of records?
    /**
     * Split size in MBs
     */
    int splitSize = 50;
    /**
     *  Number of total splits (0 for <number of tokens>, -1 for size-related generated splits) [number of tokens]
     */
    int numOfSplits = 256;

    //String command = "sudo timeout 180 ./cassandra-count -host " + host + " -keyspace ycsb -table usertable  -splitSize " + splitSize +  " -numSplits " + numOfSplits + "  2>&1";

    /**
     *  limited memory usage
     */
    //String command = "java -Xmx512m -jar cassandra-count -host " + host  + " -keyspace ycsb -table usertable  -splitSize " + splitSize +  " -numSplits " + numOfSplits + " -debug 1" ;

    String command = "./counter.sh &> count.txt && tail -n 5 count.txt" ;


    return command;
  }

  private static String getCassandraClusterStateCommand(){


    String command = "source /etc/environment && sudo -E ./cassandra/bin/nodetool status ycsb ";


    return command;
  }



  private static String getCouchbaseRecordCountCommand(){
    String commnad = "source /etc/environment && sudo -E curl -s http://$LOCAL_ADDRESS:8091/pools/default/buckets/ycsb/ -u ycsb:mowgli19  | jq '.basicStats | {itemCount: .itemCount }' | grep itemCount";

    return commnad;
  }

  private static String getCouchbaseClusterStateCommand(){
    String command = "source /etc/environment && sudo -E /opt/couchbase/bin/couchbase-cli server-list -c $LOCAL_ADDRESS --username carlos --password mowgli19";

    return command;
  }
}
