package de.uulm.omi.evaluation.task;


import static com.google.common.base.Preconditions.checkNotNull;

import de.uulm.omi.cloudiator.dbms.evaluator.model.ClusterContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoints;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationNetwork;
import de.uulm.omi.cloudiator.dbms.evaluator.model.WorkloadPhaseContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.WorkloadProcessInstance;
import de.uulm.omi.dbms.workload.api.DefaultApi;
import de.uulm.omi.utils.ColosseumUtils;
import de.uulm.omi.utils.DBMSNodeType;
import de.uulm.omi.utils.WorkloadApiClientHelper;
import java.io.File;
import java.util.List;
import org.jeasy.flows.work.DefaultWorkReport;
import org.jeasy.flows.work.WorkReport;
import org.jeasy.flows.work.WorkStatus;

/**
 * Created by Daniel Seybold on 23.01.2018.
 */
public class TSBSWorkload extends EvaluationTask {

  public enum TSBS_PHASE{
    GENERATE_DATA,
    LOAD_DATA,
    GENERATE_QUERIES,
    EXECUTE_QUERIES
  }

  private final de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSWorkload tsbsWorkload;
  private final EvaluationNetwork network;
  private final ClusterContext clusterContext;
  private final WorkloadPhaseContext workloadPhaseContext;
  private final List<String> workloadInstances;
  private final TSBS_PHASE tsbs_phase;

  private String loggerPrefix = "";

  public TSBSWorkload(EvaluationContext evaluationContext, ClusterContext clusterContext,
      de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSWorkload tsbsWorkload, EvaluationNetwork network,
      WorkloadPhaseContext workloadPhaseContext, List<String> workloadInstances, TSBS_PHASE  tsbs_phase) {
    super(TaskType.YCSB_LOAD, evaluationContext);
    this.tsbsWorkload = tsbsWorkload;
    this.network = network;
    this.clusterContext = clusterContext;
    this.workloadPhaseContext = workloadPhaseContext;
    this.workloadInstances = workloadInstances;
    this.tsbs_phase = tsbs_phase;
  }


  @Override
  public WorkReport call() {

    this.loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +": ";

    //get workloadApi client
    List<DefaultApi> workloadClients = WorkloadApiClientHelper.getWorkloadClients(workloadInstances);


    Long appInstanceId = Long.valueOf(this.clusterContext.getIdApplicationInstance());
    checkNotNull(appInstanceId);

    //build endpoints
    DBEndpoints dbEndpoints = null;

    dbEndpoints = ColosseumUtils.extractDBEndpoints(appInstanceId, this.network.getType().toString(), DBMSNodeType.ALL);


    if(dbEndpoints.isEmpty()){
      LOGGER.warn(loggerPrefix + "Found no endpoints for applicationInstanceId! Aborting!");
      return new DefaultWorkReport(WorkStatus.FAILED);
    }

    tsbsWorkload.setDbEndpoints(dbEndpoints);



    try {

      String workloadTask = "tsbs" + evaluationContext.getRunCounter().intValue() + "_" + this.evaluationContext.getScenarioType();

      int instanceCounter = 0;

      if(workloadClients.size()>1){
        LOGGER.warn("Using more than 1 workload-api instance for TSBS is currently not supported!");
        return new DefaultWorkReport(WorkStatus.FAILED);
      }

      WorkloadProcessInstance result;
      for (DefaultApi workloadClient: workloadClients) {

          switch (this.tsbs_phase){
            case GENERATE_DATA:
              writeWorkloadSpec(instanceCounter, TRANSACTION_SPEC);
              result = workloadClient.workloadTsbsGenerateDataPost(workloadTask ,tsbsWorkload);
              LOGGER.debug(loggerPrefix + this.tsbs_phase.toString() + " Phase status: " + result.getProcessStatus().getValue());
              this.workloadPhaseContext.setProcessState(result.getProcessStatus().getValue());
              break;

            case LOAD_DATA:
              result = workloadClient.workloadTsbsLoadDataPost(workloadTask ,tsbsWorkload);
              LOGGER.debug(loggerPrefix + this.tsbs_phase.toString() + " Phase status: " + result.getProcessStatus().getValue());
              this.workloadPhaseContext.setProcessState(result.getProcessStatus().getValue());
              break;

            case GENERATE_QUERIES:
              result = workloadClient.workloadTsbsGenerateQueriesPost(workloadTask ,tsbsWorkload);
              LOGGER.debug(loggerPrefix + this.tsbs_phase.toString() + " Phase status: " + result.getProcessStatus().getValue());
              this.workloadPhaseContext.setProcessState(result.getProcessStatus().getValue());
              break;

            case EXECUTE_QUERIES:
              result = workloadClient.workloadTsbsExecuteQueriesPost(workloadTask ,tsbsWorkload);
              LOGGER.debug(loggerPrefix + this.tsbs_phase.toString() + " Phase status: " + result.getProcessStatus().getValue());
              this.workloadPhaseContext.setProcessState(result.getProcessStatus().getValue());
              break;
          }


          instanceCounter++;


      }

    } catch (de.uulm.omi.dbms.workload.client.ApiException e) {
      LOGGER.error(loggerPrefix + "error calling workload API", e);

      return new DefaultWorkReport(WorkStatus.FAILED);
    }

    return new DefaultWorkReport(WorkStatus.COMPLETED);
  }

  private void writeWorkloadSpec(int instanceCounter, String workloadSpecName){

    //Write Spec to disk
    LOGGER.debug(loggerPrefix + "Running Workload Spec: " + tsbsWorkload.toString());
    String specUri = this.evaluationContext.getEvluationRun() +   File.separator + EvaluationTask.SPECS_FOLDER
        + File.separator + instanceCounter + "_" + workloadSpecName;
    this.writeReport(specUri, GSON.toJson(tsbsWorkload));
  }



}
